'use strict';

var nodeToString = require('hast-util-to-string');
var rangeParser = require('parse-numeric-range');
var refractor = require('refractor');
var rehype = require('rehype');
var parse = require('rehype-parse');
var unified = require('unified');
var unistUtilVisit = require('unist-util-visit');
var unistUtilFilter = require('unist-util-filter');
var unistUtilVisitParents = require('unist-util-visit-parents');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var nodeToString__default = /*#__PURE__*/_interopDefaultLegacy(nodeToString);
var rangeParser__default = /*#__PURE__*/_interopDefaultLegacy(rangeParser);
var refractor__default = /*#__PURE__*/_interopDefaultLegacy(refractor);
var rehype__default = /*#__PURE__*/_interopDefaultLegacy(rehype);
var parse__default = /*#__PURE__*/_interopDefaultLegacy(parse);
var unified__default = /*#__PURE__*/_interopDefaultLegacy(unified);

/**
 * Weird "map" polyfill, that's weak if possible
 */

const NodeMap = typeof WeakMap === 'function' ? HappyMap : SadMap;

// Happy path! Fakes a `clear()` if possible (noop)
function HappyMap() {
  this.map = new WeakMap();
}

HappyMap.prototype.has = function (key) {
  return this.map.has(key);
};

HappyMap.prototype.set = function (key, value) {
  this.map.set(key, value);
  return this;
};

HappyMap.prototype.get = function (key) {
  return this.map.get(key);
};

HappyMap.prototype.clear = function () {
  // intentional noop, since we dont need to/cant clear
};

// Sad path! Use less stylish approach
function SadMap() {
  this.keys = [];
  this.values = [];
}

SadMap.prototype.has = function (key) {
  return this.keys.indexOf(key) !== -1;
};

SadMap.prototype.set = function (key, value) {
  const index = this.keys.indexOf(key);
  if (index === -1) {
    this.keys.push(key);
    this.values.push(value);
  } else {
    this.values[index] = value;
  }
  return this;
};

SadMap.prototype.get = function (key) {
  const index = this.keys.indexOf(key);
  return index === -1 ? undefined : this.values[index];
};

SadMap.prototype.clear = function () {
  this.keys = [];
  this.values = [];
};

// Copied from https://github.com/rexxars/react-refractor/blob/master/src/Refractor.js

function addMarkers(ast, options) {
  const markers = options.markers
    .map((marker) => (marker.line ? marker : { line: marker }))
    .sort((nodeA, nodeB) => nodeA.line - nodeB.line);

  const numbered = lineNumberify(ast).nodes;
  return wrapLines(numbered, markers, options);
}

function lineNumberify(ast, context = { lineNumber: 1 }) {
  return ast.reduce(
    (result, node) => {
      const lineStart = context.lineNumber;

      if (node.type === 'text') {
        if (node.value.indexOf('\n') === -1) {
          node.lineStart = lineStart;
          node.lineEnd = lineStart;
          result.nodes.push(node);
          return result;
        }

        const lines = node.value.split('\n');
        for (let i = 0; i < lines.length; i++) {
          const lineNum = i === 0 ? context.lineNumber : ++context.lineNumber;
          result.nodes.push({
            type: 'text',
            value: i === lines.length - 1 ? lines[i] : `${lines[i]}\n`,
            lineStart: lineNum,
            lineEnd: lineNum,
          });
        }

        result.lineNumber = context.lineNumber;
        return result;
      }

      if (node.children) {
        const processed = lineNumberify(node.children, context);
        const firstChild = processed.nodes[0];
        const lastChild = processed.nodes[processed.nodes.length - 1];
        node.lineStart = firstChild ? firstChild.lineStart : lineStart;
        node.lineEnd = lastChild ? lastChild.lineEnd : lineStart;
        node.children = processed.nodes;
        result.lineNumber = processed.lineNumber;
        result.nodes.push(node);
        return result;
      }

      result.nodes.push(node);
      return result;
    },
    { nodes: [], lineNumber: context.lineNumber },
  );
}

function unwrapLine(markerLine, nodes) {
  const tree = { type: 'root', children: nodes };

  const headMap = new NodeMap();
  const lineMap = new NodeMap();
  const tailMap = new NodeMap();
  const cloned = [];

  function addCopy(map, node, ancestors) {
    cloned.push(node);

    ancestors.forEach((ancestor) => {
      if (!map.has(ancestor)) {
        map.set(ancestor, Object.assign({}, ancestor, { children: [] }));

        if (ancestor !== tree) {
          cloned.push(ancestor);
        }
      }
    });

    let i = ancestors.length;
    while (i--) {
      const ancestor = map.get(ancestors[i]);
      const child = ancestors[i + 1];
      const leaf = map.get(child) || node;
      if (ancestor.children.indexOf(leaf) === -1) {
        ancestor.children.push(leaf);
      }
    }
  }

  unistUtilVisitParents.visitParents(tree, (node, ancestors) => {
    if (node.children) {
      return;
    }

    // These nodes are on previous lines, but nested within the same structure
    if (node.lineStart < markerLine) {
      addCopy(headMap, node, ancestors);
      return;
    }

    // These nodes are on the target line
    if (node.lineStart === markerLine) {
      addCopy(lineMap, node, ancestors);
      return;
    }

    // If we have shared ancestors with some of the cloned elements,
    // create another tree of the remaining nodes
    if (
      node.lineEnd > markerLine &&
      cloned.some((clone) => ancestors.indexOf(clone) !== -1)
    ) {
      addCopy(tailMap, node, ancestors);
    }
  });

  // Get the remaining nodes - the ones who were not part of the same tree
  const filtered = unistUtilFilter.filter(tree, (node) => cloned.indexOf(node) === -1);
  const getChildren = (map) => {
    const rootNode = map.get(tree);
    if (!rootNode) {
      return [];
    }

    unistUtilVisitParents.visitParents(rootNode, (leaf, ancestors) => {
      if (leaf.children) {
        leaf.lineStart = 0;
        leaf.lineEnd = 0;
        return;
      }

      ancestors.forEach((ancestor) => {
        ancestor.lineStart = Math.max(ancestor.lineStart, leaf.lineStart);
        ancestor.lineEnd = Math.max(ancestor.lineEnd, leaf.lineEnd);
      });
    });

    return rootNode.children;
  };

  const merged = [].concat(
    getChildren(headMap),
    getChildren(lineMap),
    getChildren(tailMap),
    filtered ? filtered.children : [],
  );

  headMap.clear();
  lineMap.clear();
  tailMap.clear();

  return merged;
}

function wrapBatch(children, marker, options) {
  const className = marker.className || 'mdx-marker';
  return {
    type: 'element',
    tagName: marker.component || 'div',
    properties: marker.component
      ? Object.assign({}, options, { className })
      : { className },
    children,
    lineStart: marker.line,
    lineEnd: children[children.length - 1].lineEnd,
    isMarker: true,
  };
}

function wrapLines(treeNodes, markers, options) {
  if (markers.length === 0 || treeNodes.length === 0) {
    return treeNodes;
  }

  const ast = markers.reduce(
    (acc, marker) => unwrapLine(marker.line, acc),
    treeNodes,
  );

  // Container for the new AST
  const wrapped = [];

  // Note: Markers are already sorted by line number (ascending)
  let astIndex = 0;
  for (let m = 0; m < markers.length; m++) {
    const marker = markers[m];

    // Start by eating all AST nodes with line numbers up to the given marker
    for (
      let node = ast[astIndex];
      node && node.lineEnd < marker.line;
      node = ast[++astIndex]
    ) {
      wrapped.push(node);
    }

    // Now proceed to find all _contiguous_ nodes on the same line
    const batch = [];
    for (
      let node = ast[astIndex];
      node && node.lineEnd === marker.line;
      node = ast[++astIndex]
    ) {
      batch.push(node);
    }

    // Now add that batch, if we have anything
    if (batch.length > 0) {
      wrapped.push(wrapBatch(batch, marker, options));
    }
  }

  // Now add the remaining AST nodes
  while (astIndex < ast.length) {
    wrapped.push(ast[astIndex++]);
  }

  return wrapped;
}

/**
 * This module walks through the node tree and does:
 * - gets the class name
 * - parses the class and extracts the highlight lines directive and the language name
 * - highlights the code using refractor
 * - if markers are present then:
 *    - converts AST to HTML
 *    - then applies some fixes to make line highlighting work with JSX found here: https://github.com/gatsbyjs/gatsby/blob/master/packages/gatsby-remark-prismjs/src/directives.js#L113-L119
 *    - add markers using: https://github.com/rexxars/react-refractor/blob/master/src/addMarkers.js
 *    - converts the code back from HTML to AST
 *  - sets the code as value
 */

function mdxPrism2(options = {}) {
  return (tree) => {
    unistUtilVisit.visit(tree, 'element', visitor);
  };

  function visitor(node, index, parent) {
    if (!parent || parent.tagName !== 'pre' || node.tagName !== 'code') {
      return;
    }

    const className = getLangClass(node);
    const { highlightLines, splitLanguage } = parseLineNumberRange(className);
    const lang = getLanguage(splitLanguage);
    const markers = highlightLines;

    if (lang === null) {
      return;
    }

    let result;

    try {
      parent.properties.className = (parent.properties.className || []).concat(
        'language-' + lang,
      );

      result = refractor__default['default'].highlight(nodeToString__default['default'](node), lang);

      if (markers && markers.length > 0) {
        // This blocks attempts this fix:
        // https://github.com/gatsbyjs/gatsby/blob/master/packages/gatsby-remark-prismjs/src/directives.js#L113-L119
        const PLAIN_TEXT_WITH_LF_TEST = /<span class="token plain-text">[^<]*\n[^<]*<\/span>/g;

        // AST to HTML
        let html_ = rehype__default['default']()
          .stringify({ type: 'root', children: result })
          .toString();

        // Fix JSX issue
        html_ = html_.replace(PLAIN_TEXT_WITH_LF_TEST, (match) => {
          return match.replace(
            /\n/g,
            '</span>\n<span class="token plain-text">',
          );
        });

        // HTML to AST
        const hast_ = unified__default['default']()
          .use(parse__default['default'], { emitParseErrors: true, fragment: true })
          .parse(html_);

        // Add markers
        result = addMarkers(hast_.children, { markers });
      }
    } catch (err) {
      if (options.ignoreMissing && /Unknown language/.test(err.message)) {
        return;
      }

      throw err;
    }

    node.children = result;
  }
}

const parseLineNumberRange = (language) => {
  if (!language) {
    return '';
  }

  const lineRegex = /{.*}/g;

  if (lineRegex.test(language)) {
    let [splitLanguage, ...options] = language.split('{');

    let highlightLines = [];

    options.forEach((option) => {
      option = option.slice(0, -1);
      if (rangeParser__default['default'](option).length > 0) {
        highlightLines = rangeParser__default['default'](option).filter((n) => n > 0);
      }
    });

    const result = {
      splitLanguage,
      highlightLines,
    };

    return result;
  }

  return { splitLanguage: language };
};

function getLangClass(node) {
  const className = node.properties.className || [];
  for (const item of className) {
    if (item.slice(0, 9) === 'language-') {
      return item;
    }
  }
  return null;
}

function getLanguage(className = '') {
  if (className.slice(0, 9) === 'language-') {
    return className.slice(9).toLowerCase();
  }

  return null;
}

module.exports = mdxPrism2;
//# sourceMappingURL=index.cjs.js.map
